;;; Directory Local Variables            -*- no-byte-compile: t -*-
;;; For more information see (info "(emacs) Directory Variables")

((nil . ((indent-tabs-mode . nil)
         (sentence-end-double-space . t)
         (fill-column . 80)
         (ispell-buffer-session-localwords . ("transclusion")))))
